# borinud/db.py - database.
#
# Copyright (C) 2013 ARPA-SIM <urpsim@smr.arpa.emr.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Emanuele Di Giacomo <edigiacomo@arpa.emr.it>

import dballe
import json

class DB(object):
    """Abstract class.
    A concrete DB must implement the following methods
    - query_stations
    - query_summary
    - query_data
    """

    @classmethod
    def get(cls, url, *args, **kw):
        """Factory method.

            Keyword arguments
        - cached_summary: filename of the cached summary
        """
        db = None
        if url.startswith("sqlite:") or url.startswith("odbc:"):
            db = DballeDB(url)
        else:
            raise Exception("Invalid db url: " + url)

        if kw.get("cached_summary"):
            db = SummaryCacheDB(db, kw.get("cached_summary"))

        return db

    def query_stations(self, rec):
        """Query stations. Return a dballe.Record."""
        raise NotImplementedError()

    def query_summary(self, rec):
        """Query summary. Return a dballe.Record."""
        raise NotImplementedError()

    def query_data(self, rec):
        """Query data. Return a dballe.Record."""
        raise NotImplementedError()


class DballeDB(DB):
    """DB-All.e database."""
    def __init__(self, url):
        """Create a DB-All.e database from `url` DSN."""
        self.url = url

    def __open_db(self):
        """Open the database."""
        return dballe.DB.connect_from_url(self.url)

    def query_stations(self, rec):
        db = self.__open_db()
        rec.set_station_context()
        return db.query_data(rec)

    def query_summary(self, rec):
        db = self.__open_db()
        return db.query_summary(rec)

    def query_data(self, rec):
        db = self.__open_db()
        return db.query_data(rec)


class SummaryCacheDB(DB):
    """Preemptive summary cache.

    The cache must be loaded and updated using the method
    `write_cached_summary`.
    For example, a crontab script can update every 10 minutes the cache file:

        # cacheupdater.py
        from borinud.db import SummaryCacheDB, DballeDB
        c = SummaryCacheDB(DballeDB("sqlite:mydb.sqlite"), "/tmp/cache.json")
        c.write_cached_summary()

        # crontab
        */10 * * * * python cacheupdater.py
    """
    def __init__(self, db, cachefile, ttl=None):
        """Creates a summary cache for the database `db` reading the file with
        name `cachefile`.

        The summary cache can be loaded in memory setting the parameter `ttl`
        (number of seconds the memory cache lives).
        """
        self.db = db
        self.cachefile = cachefile
        self.ttl = ttl

    def update_expirydate(self):
        """Update the expirydate of the in-memory cache.

        If `ttl` is None, do nothing.
        """
        import datetime
        if self.ttl is not None:
            self.expirydate = datetime.datetime.now() + datetime.timedelta(self.ttl)

    def is_expired(self):
        """Return True if the in-memory cache is expired, False otherwise.

        If `ttl` is None, this method always return True
        """
        if getattr(self, "expirydate", None) is None:
            return True
        else:
            import datetime
            return self.expirydate < datetime.datetime.now()

    def get_cached_summary(self):
        """Get the cached summary.

        If the in-memory cache is expired, read the cache from the file and
        update the expiry date.
        """
        if self.is_expired():
            self.summary = self.read_cached_summary()
            self.update_expirydate()
        return self.summary

    def read_cached_summary(self):
        """Read the summary from the cache file."""
        from .codec import SummaryJSONDecoder
        with open(self.cachefile) as f:
            return json.load(f, cls=SummaryJSONDecoder)

    def write_cached_summary(self):
        """Write the db summary to the cache file."""
        import os
        from tempfile import NamedTemporaryFile
        # The summary is first written in a temporary file and then moved to the
        # right path (os.rename is atomic in POSIX OS)
        cachedir = os.path.realpath(os.path.dirname(self.cachefile))
        with NamedTemporaryFile(delete=False, dir=cachedir) as f:
            try:
                from .codec import SummaryJSONEncoder
                json.dump(
                    self.db.query_summary(dballe.Record()),
                    f,
                    cls=SummaryJSONEncoder
                )
                # Atomic rename in POSIX OS
                os.rename(f.name, self.cachefile)
            except:
                os.unlink(f.name)
                raise

    def get_filter_summary(self, rec):
        """Return a filter function based on dballe.Record `rec`.

        The following keys are considered
        - ident
        - lon
        - lat
        - rep_memo
        - trange
        - level
        - var
        """
        def wrapper(item):
            if rec.get("ident") and rec.get("ident") != item.get("ident"):
                return False
            elif rec.get("lon") and rec.key("lon") != item.key("lon"):
                return False
            elif rec.get("lat") and rec.key("lat") != item.key("lat"):
                return False
            elif rec.get("rep_memo") and rec.get("rep_memo") != item.get("rep_memo"):
                return False
            elif rec.get("trange") and rec.get("trange") != item.get("trange"):
                return False
            elif rec.get("level") and rec.get("level") != item.get("level"):
                return False
            elif rec.get("var") and rec.get("var") != item.get("var"):
                return False
            return True
        return wrapper

    def query_stations(self, rec):
        return self.db.query_stations(rec)

    def query_summary(self, rec):
        return filter(
            self.get_filter_summary(rec),
            self.get_cached_summary()
        )

    def query_data(self, rec):
        return self.db.query_data(rec)



# borinud/codec.py - encoding/decoding utilities
#
# Copyright (C) 2013 ARPA-SIM <urpsim@smr.arpa.emr.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Emanuele Di Giacomo <edigiacomo@arpa.emr.it>
import json
import dballe

class BaseJSONEncoder(json.JSONEncoder):
    """Base JSON encoder."""
    def default(self, o):
        from datetime import datetime
        if isinstance(o, datetime):
            return o.isoformat()
        else:
            return super(BaseJSONEncoder, self).default(o)

class GeoJSONEncoder(BaseJSONEncoder):
    """GeoJSON encoder."""
    def encode(self, o):
        try:
            iterable = iter(o)
        except TypeError:
            return super(GeoJSONEncoder, self).encode(o)
        else:
            return super(GeoJSONEncoder, self).encode(self.toFeatureCollection(o))

    def default(self, o):
        from datetime import datetime
        if isinstance(o, datetime) and o == datetime(1000, 1, 1, 0, 0, 0):
                return None
        else:
            return super(GeoJSONEncoder, self).default(o)

    def toSkip(self, record):
        """True if the record must be skipped, false otherwise."""
        from datetime import datetime
        if record.get("date") == datetime(1000, 1, 1, 0, 0, 0) and record.get("var") in ( "B05001", "B06001", "B01194" ):
            return True
        else:
            return False

    def toFeatureCollection(self, cursor):
        """Convert a collection of items to a feature collection."""
        return {
            "type": "FeatureCollection",
            "features": [ self.toFeature(r) for r in cursor if not self.toSkip(r) ]
        }

    def toFeature(self, rec):
        """Convert a record to a feature."""
        return {
            "type": "Feature",
            "geometry": {
                "type": "Point",
                "coordinates": [ rec["lon"], rec["lat"] ],
            },
            "properties": self.toProperties(rec)
        }

    def toProperties(self, rec):
        """Get feature properties from a record."""
        p = {
            "ident": rec.get("ident"),
            "lon": rec.key("lon").enqi(),
            "lat": rec.key("lat").enqi(),
            "network": rec["rep_memo"],
            "level_t1": rec["level"][0],
            "level_v1": rec["level"][1],
            "level_t2": rec["level"][2],
            "level_v2": rec["level"][3],
            "trange_pind": rec["trange"][0],
            "trange_p1": rec["trange"][1],
            "trange_p2": rec["trange"][2],
            "bcode": rec["var"],
            "datetime": None
        }
        if rec.get("date"):
            p["datetime"] = rec.get("date")
            p["value"] = rec.get(rec["var"])
        elif rec.date_extremes() != (None, None):
            p["datetime"] = rec.date_extremes()
        return p

class SummaryJSONEncoder(BaseJSONEncoder):
    """Encode a JSON summary."""
    def default(self, o):
        if isinstance(o, dballe.Cursor):
            return [ r.copy() for r in o ]
        elif isinstance(o, dballe.Record):
            return {
                "ident": o.get("ident"),
                "lon": o.key("lon").enqi(),
                "lat": o.key("lat").enqi(),
                "rep_memo": o.get("rep_memo"),
                "level": o.get("level"),
                "trange": o.get("trange"),
                "bcode": o.get("var"),
                "date": o.date_extremes(),
            }
        else:
            return super(SummaryJSONEncoder, self).default(o)

class SummaryJSONDecoder(json.JSONDecoder):
    """Decode a JSON summary."""
    def decode(self, s):
        # TODO: decode datemin and datemax
        jsonsumm = super(SummaryJSONDecoder, self).decode(s)
        return tuple(dballe.Record(**{
            "ident": None if i["ident"] is None else i["ident"].encode(),
            "lon": i["lon"],
            "lat": i["lat"],
            "rep_memo": i["rep_memo"].encode(),
            "level": i["level"],
            "trange": i["trange"],
            "var": i["bcode"].encode(),
        }) for i in jsonsumm)



import os
from setuptools import setup, find_packages


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

name = "borinud"
version = "0.7"
release = version + ".2"

setup(
    name=name,
    version=release,
    author="Emanuele Di Giacomo",
    author_email="edigiacomo@arpa.emr.it",
    description="weather data web api",
    long_description="\n".join((
        read("README.txt"),
        read("CHANGES.txt"),
        read("CONTRIBUTORS.txt"),
    )),
    license="GPLv2+",
    keywords="ARPA SIMC weather station sensor GIS",
    url="http://www.arpa.emr.it/simc",
    packages=find_packages(),
    namespace_packages=[],
    classifiers=[
    ],
    test_suite="borinud",
    install_requires=[
        'bottle',
    ],
    extras_require={
    },
    command_options={
        'build_sphinx': {
            'project': ('setup.py', name),
            'version': ('setup.py', version),
            'release': ('setup.py', release),
            'source_dir': ('setup.py', 'doc/source'),
            'build_dir': ('setup.py', 'doc/build'),
        }
    },
    scripts=[
    ],
    include_package_data=True,
)



# borinud/ws.py - web services
#
# Copyright (C) 2013 ARPA-SIM <urpsim@smr.arpa.emr.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Emanuele Di Giacomo <edigiacomo@arpa.emr.it>
import bottle
import dballe
import json
from borinud import __version__

__apiversion__ = "0.1"

class SerializerPlugin(object):
    name = "borinud-serialize"
    api = 2

    def setup(self, app):
        app.uninstall('json')

    def togeojson(self, response):
        from borinud.codec import GeoJSONEncoder
        response = json.dumps(response, cls=GeoJSONEncoder)
        jsonp = bottle.request.query.get("callback")
        if jsonp:
            bottle.response.content_type = "application/javascript"
            response = jsonp + "(" + response + ")"
        else:
            bottle.response.content_type = "application/json"
        return response

    def apply(self, callback, context):
        def wrapper(*args, **kwargs):
            response = callback(*args, **kwargs)
            if isinstance(response, tuple) or isinstance(response, dballe.Cursor):
                fmt = bottle.request.query.get("format", "geojson")
                if fmt == "geojson":
                    return self.togeojson(response)
                else:
                    raise Exception("Unsupported format {0}".format(fmt))
            else:
                return response
        return wrapper

app = bottle.Bottle(autojson=False)
app.install(SerializerPlugin())

@app.hook("after_request")
def enable_cors():
    bottle.response.headers["Access-Control-Allow-Origin"] = app.config.get("cors_site")


@app.get("")
@app.get("/")
def get_index():
    return get_static_file("index.html")

@app.get("/static/<filename:path>")
def get_static_file(filename):
    from pkg_resources import Requirement, resource_filename
    rootdir = resource_filename(Requirement.parse("borinud"), "borinud/_static")
    return bottle.static_file(filename, root=rootdir)

@app.get("/apiversion")
def get_api_version():
    bottle.response.content_type = "text/plain"
    return __apiversion__

# Return the summary
@app.get("/*/*/*/*/*/*/summaries")
@app.get("/*/*/<network>/*/*/*/summaries")
@app.get("/-/<lon:int>,<lat:int>/<network>/*/*/*/summaries")
@app.get("/<ident:re:[^*-]+>/*/<network>/*/*/*/summaries")
@app.get("/*/*/<network>/<pind:int>,<p1:int>,<p2:int>/<lt1:int>,<l1>,<lt2>,<l2>/<bcode>/summaries/<year:int>/<month:int>")
@app.get("/*/*/<network>/<pind:int>,<p1:int>,<p2:int>/<lt1:int>,<l1>,<lt2>,<l2>/<bcode>/summaries/<year:int>/<month:int>/<day:int>")
def get_summaries(ident=None, network=None, lon=None, lat=None,
                  pind=None, p1=None, p2=None,
                  lt1=None, l1=None, lt2=None, l2=None, bcode=None,
                  year=None, month=None, day=None):
    query = dballe.Record()
    query["ident"] = ident
    query["lon"] = lon
    query["lat"] = lat
    query["rep_memo"] = network
    query["trange"] = ( pind, p1, p2 )

    for n, v in (
        ("leveltype1", lt1),
        ("l1", l1),
        ("leveltype2", lt2),
        ("l2", l2),
    ):
        if v is not None:
            query.set_from_string("%s=%s" % (n, v))

    query["var"] = bcode
    query["year"] = year
    query["month"] = month
    query["day"] = day
    return app.config.db.query_summary(query)

# Timeseries
@app.get("/<ident>/<lon:int>,<lat:int>/<network>/<pind:int>,<p1:int>,<p2:int>/<lt1:int>,<l1>,<lt2>,<l2>/<bcode>/timeseries/<year:int>")
@app.get("/<ident>/<lon:int>,<lat:int>/<network>/<pind:int>,<p1:int>,<p2:int>/<lt1:int>,<l1>,<lt2>,<l2>/<bcode>/timeseries/<year:int>/<month:int>")
@app.get("/<ident>/<lon:int>,<lat:int>/<network>/<pind:int>,<p1:int>,<p2:int>/<lt1:int>,<l1>,<lt2>,<l2>/<bcode>/timeseries/<year:int>/<month:int>/<day:int>")
def get_resource_timeseries(ident,
                            lon, lat, network,
                            pind, p1, p2,
                            lt1, l1, lt2, l2,
                            bcode,
                            year, month=None, day=None):
    query = dballe.Record()
    if ident != "-": query["ident"] = ident
    query["lon"] = lon
    query["lat"] = lat
    query["rep_memo"] = network
    query["trange"] = ( pind, p1, p2 )
    query.set_from_string("leveltype1=%s" % (lt1,))
    query.set_from_string("l1=%s" % (l1,))
    query.set_from_string("leveltype2=%s" % (lt2,))
    query.set_from_string("l2=%s" % (l2,))
    query["var"] = bcode
    query["year"] = year
    query["month"] = month
    query["day"] = day

    return app.config.db.query_data(query)

# Spatialseries
@app.get("/*/*/<network>/<pind:int>,<p1:int>,<p2:int>/<lt1:int>,<l1>,<lt2>,<l2>/<bcode>/spatialseries/<year:int>/<month:int>/<day:int>/<hour:int>")
def get_network_spatialseries(network,
                              pind, p1, p2,
                              lt1, l1, lt2, l2,
                              bcode,
                              year, month, day, hour):
    from datetime import datetime, timedelta
    query = dballe.Record()
    query["rep_memo"] = network
    query["trange"] = ( pind, p1, p2 )
    query.set_from_string("leveltype1=%s" % (lt1,))
    query.set_from_string("l1=%s" % (l1,))
    query.set_from_string("leveltype2=%s" % (lt2,))
    query.set_from_string("l2=%s" % (l2,))
    query["var"] = bcode
    d = datetime(year, month, day, hour)
    query["datemin"] = d - timedelta(seconds=1800)
    query["datemax"] = d + timedelta(seconds=1799)

    return app.config.db.query_data(query)

# Station data
@app.get("/-/<lon:int>,<lat:int>/<network>/-,-,-/257,-,-,-/*/stationdata")
@app.get("<ident>/*/<network>/-,-,-/257,-,-,-/*/stationdata")
def get_station_data(network, ident=None, lon=None, lat=None):
    query = dballe.Record()
    query.set_station_context()
    query["rep_memo"] = network
    query["ident"] = ident
    query["lon"] = lon
    query["lat"] = lat

    return app.config.db.query_stations(query)

# Get local B table
@app.get("/descriptions/btable")
def get_btable():
    vartable = dballe.Vartable.get("dballe")

    return dict((b.var, {
        "description": b.desc,
        "unit": b.unit
    }) for b in vartable)

versionedapp = bottle.Bottle()

@versionedapp.get("/apiversion")
def get_versioned_api_version():
    bottle.response.content_type = "text/plain"
    return __apiversion__

versionedapp.mount(__apiversion__, app)

if __name__ == '__main__':
    from argparse import ArgumentParser
    parser = ArgumentParser(description="ObsWs Web Server")
    parser.add_argument("--version",
                        action="version",
                        version="%(prog)s " + __version__)
    parser.add_argument("--apiversion",
                        action="version",
                        help="show web api version number and exit",
                        version="%(prog)s WS API " + __apiversion__)
    parser.add_argument("--dsn", metavar="URL",
                        help="url for dballe db",
                        required=True)
    parser.add_argument("--host", metavar="HOST",
                        help="host name",
                        default="localhost")
    parser.add_argument("--port", metavar="PORT",
                        help="port",
                        default="8080")
    parser.add_argument("--server", metavar="SERVER",
                        help="server name",
                        default='wsgiref')
    parser.add_argument("--allow-cors", metavar="SITE",
                        help="allow CORS requests from site")
    parser.add_argument("--versioned", action="store_true",
                        help="Use versioned API",
                        default=False)
    parser.add_argument("--cached-summary", metavar="FILE",
                        help="cached summary")
    opts = parser.parse_args()
    from borinud.db import DB
    app.config.db = DB.get(
        opts.dsn,
        cached_summary=opts.cached_summary
    )
    app.config.cors_site = opts.allow_cors

    if opts.versioned:
        apptorun = versionedapp
    else:
        apptorun = app

    bottle.run(app=apptorun,
               host=opts.host,
               port=opts.port,
               server=opts.server)
